from datetime import datetime

from django.db import models

from django.conf import settings
from django.contrib.contenttypes.generic import GenericForeignKey
from django.contrib.contenttypes.models import ContentType


class Comment(models.Model):

    author = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, related_name="comments")

    name = models.CharField(max_length=100)
    email = models.CharField(max_length=255, blank=True)
    website = models.CharField(max_length=255, blank=True)

    content_type = models.ForeignKey(ContentType)
    object_id = models.IntegerField()
    content_object = GenericForeignKey()

    comment = models.TextField()

    submit_date = models.DateTimeField(default=datetime.now)
    ip_address = models.GenericIPAddressField(null=True)
    public = models.BooleanField(default=True)

    def __unicode__(self):
        return "pk=%d" % self.pk



from django import template
from django.core.urlresolvers import reverse

from django.contrib.contenttypes.models import ContentType

from dialogos.authorization import load_can_delete, load_can_edit
from dialogos.forms import CommentForm
from dialogos.models import Comment


can_delete = load_can_delete()
can_edit = load_can_edit()
register = template.Library()


@register.filter
def can_edit_comment(comment, user):
    return can_edit(user, comment)


@register.filter
def can_delete_comment(comment, user):
    return can_delete(user, comment)


class BaseCommentNode(template.Node):
    
    @classmethod
    def handle_token(cls, parser, token):
        bits = token.split_contents()
        if not cls.requires_as_var and len(bits) == 2:
            return cls(parser.compile_filter(bits[1]))
        elif len(bits) == 4:
            if bits[2] != "as":
                raise template.TemplateSyntaxError("%r's 2nd argument must be 'as'" % bits[0])
            return cls(parser.compile_filter(bits[1]), bits[3])
        if cls.requires_as_var:
            args = "1 argument"
        else:
            args = "either 1 or 3 arguments"
        raise template.TemplateSyntaxError("%r takes %s" % (bits[0], args))
    
    def __init__(self, obj, varname=None):
        self.obj = obj
        self.varname = varname
    
    def get_comments(self, context):
        obj = self.obj.resolve(context)
        comments = Comment.objects.filter(
            object_id=obj.pk,
            content_type=ContentType.objects.get_for_model(obj)
        )
        return comments.order_by("submit_date")


class CommentCountNode(BaseCommentNode):
    
    requires_as_var = False
    
    def render(self, context):
        comments = self.get_comments(context).count()
        if self.varname is not None:
            context[self.varname] = comments
            return ""
        return unicode(comments)


class CommentsNode(BaseCommentNode):
    
    requires_as_var = True
    
    def render(self, context):
        context[self.varname] = self.get_comments(context)
        return ""


class CommentFormNode(BaseCommentNode):
    
    requires_as_var = False
    
    def render(self, context):
        obj = self.obj.resolve(context)
        user = context.get("user")
        form_class = context.get("form", CommentForm)
        form = form_class(obj=obj, user=user)
        context[self.varname] = form
        return ""


class CommentTargetNode(BaseCommentNode):
    
    requires_as_var = False
    
    def render(self, context):
        obj = self.obj.resolve(context)
        return reverse("post_comment", kwargs={
            "content_type_id": ContentType.objects.get_for_model(obj).pk,
            "object_id": obj.pk
        })


@register.tag
def comment_count(parser, token):
    """
    Usage:
        
        {% comment_count obj %}
    or
        {% comment_count obj as var %}
    """
    return CommentCountNode.handle_token(parser, token)


@register.tag
def comments(parser, token):
    """
    Usage:
        
        {% comments obj as var %}
    """
    return CommentsNode.handle_token(parser, token)


@register.tag
def comment_form(parser, token):
    """
    Usage:
        
        {% comment_form obj as comment_form %}
        
    Will read the `user` var out of the contex to know if the form should be
    form an auth'd user or not.
    """
    return CommentFormNode.handle_token(parser, token)


@register.tag
def comment_target(parser, token):
    """
    Usage:
        
        {% comment_target obj [as varname] %}
    """
    return CommentTargetNode.handle_token(parser, token)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
try:
    from django.utils.importlib import import_module
except ImportError:
    from importlib import import_module


def load_path_attr(path):
    i = path.rfind(".")
    module, attr = path[:i], path[i + 1:]
    try:
        mod = import_module(module)
    except ImportError, e:
        raise ImproperlyConfigured("Error importing %s: '%s'" % (module, e))
    try:
        attr = getattr(mod, attr)
    except AttributeError:
        raise ImproperlyConfigured("Module '%s' does not define a '%s'" % (module, attr))
    return attr


def default_can_delete(user, comment):
    if user.is_superuser:
        return True
    return user == comment.author


def default_can_edit(user, comment):
    return user == comment.author


def load_can_delete():
    import_path = getattr(settings, "COMMENTS_CAN_DELETE_CALLABLE", None)
    
    if import_path is None:
        return default_can_delete
    
    return load_path_attr(import_path)


def load_can_edit():
    import_path = getattr(settings, "COMMENTS_CAN_EDIT_CALLABLE", None)
    
    if import_path is None:
        return default_can_edit
    
    return load_path_attr(import_path)



from django import forms

from django.contrib.contenttypes.models import ContentType

from dialogos.models import Comment


class CommentForm(forms.ModelForm):
    
    class Meta:
        model = Comment
        fields = [
            "name", "email", "website", "comment"
        ]
    
    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request", None)
        self.obj = kwargs.pop("obj")
        self.user = kwargs.pop("user")
        super(CommentForm, self).__init__(*args, **kwargs)
        if self.user is not None and not self.user.is_anonymous():
            del self.fields["name"]
            del self.fields["email"]
            del self.fields["website"]
    
    def save(self, commit=True):
        comment = super(CommentForm, self).save(commit=False)
        comment.ip_address = self.request.META.get("REMOTE_ADDR", None)
        comment.content_type = ContentType.objects.get_for_model(self.obj)
        comment.object_id = self.obj.pk
        if self.user is not None and not self.user.is_anonymous():
            comment.author = self.user
        if commit:
            comment.save()
        return comment



from django.core.urlresolvers import reverse
from django.template import Template, Context
from django.test import TestCase

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType

from dialogos.forms import CommentForm
from dialogos.models import Comment


class login(object):
    def __init__(self, testcase, user, password):
        self.testcase = testcase
        success = testcase.client.login(username=user, password=password)
        self.testcase.assertTrue(
            success,
            "login with username=%r, password=%r failed" % (user, password)
        )
    
    def __enter__(self):
        pass
    
    def __exit__(self, *args):
        self.testcase.client.logout()


class TestCaseMixin(object):
    def get(self, url_name, *args, **kwargs):
        data = kwargs.pop("data", {})
        return self.client.get(reverse(url_name, args=args, kwargs=kwargs), data)
    
    def getajax(self, url_name, *args, **kwargs):
        data = kwargs.pop("data", {})
        return self.client.get(reverse(url_name, args=args, kwargs=kwargs), data,
                               HTTP_X_REQUESTED_WITH="XMLHttpRequest")
    
    def post(self, url_name, *args, **kwargs):
        data = kwargs.pop("data", {})
        return self.client.post(reverse(url_name, args=args, kwargs=kwargs), data)
    
    def postajax(self, url_name, *args, **kwargs):
        data = kwargs.pop("data", {})
        return self.client.post(reverse(url_name, args=args, kwargs=kwargs), data,
                                HTTP_X_REQUESTED_WITH="XMLHttpRequest")
    
    def login(self, user, password):
        return login(self, user, password)
    
    def reload(self, obj):
        return obj.__class__._default_manager.get(pk=obj.pk)
    
    def assert_renders(self, tmpl, context, value):
        tmpl = Template(tmpl)
        self.assertEqual(tmpl.render(context), value)


class CommentTests(TestCaseMixin, TestCase):
    
    def setUp(self):
        self.user = get_user_model().objects.create_user("gimli", "myaxe@dwarf.org", "gloin")
        self.user2 = get_user_model().objects.create_user("aragorn", "theking@gondor.gov", "strider")
    
    def assert_renders(self, tmpl, context, value):
        tmpl = Template(tmpl)
        self.assertEqual(tmpl.render(context), value)
    
    def post_comment(self, obj, data):
        return self.post("post_comment",
            content_type_id=ContentType.objects.get_for_model(obj).pk,
            object_id=obj.pk,
            data=data
        )
    
    def test_post_comment(self):
        g = get_user_model().objects.create(username="Gandalf")
        
        response = self.post_comment(g, data={
            "name": "Frodo Baggins",
            "comment": "Where'd you go?",
        })
        self.assertEqual(response.status_code, 302)
        
        self.assertEqual(Comment.objects.count(), 1)
        c = Comment.objects.get()
        self.assertEqual(c.author, None)
        self.assertEqual(c.name, "Frodo Baggins")
        
        response = self.post_comment(g, data={
            "comment": "Where is everyone?"
        })
        self.assertEqual(Comment.objects.count(), 1)
        
        with self.login("gimli", "gloin"):
            response = self.post_comment(g, data={
                "comment": "I thought you were watching the hobbits?"
            })
            self.assertEqual(response.status_code, 302)
            self.assertEqual(Comment.objects.count(), 2)
            
            c = Comment.objects.order_by("id")[1]
            self.assertEqual(c.comment, "I thought you were watching the hobbits?")
            self.assertEqual(c.author, self.user)
    
    def test_delete_comment(self):
        g = get_user_model().objects.create(username="Boromir")
        with self.login("gimli", "gloin"):
            response = self.post_comment(g, data={
                "comment": "Wow, you're a jerk.",
            })
            comment = Comment.objects.get()
        
        response = self.post("delete_comment", comment_id=comment.pk)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(Comment.objects.count(), 1)
        
        with self.login("aragorn", "strider"):
            response = self.post("delete_comment", comment_id=comment.pk)
            self.assertEqual(response.status_code, 302)
            self.assertEqual(Comment.objects.count(), 1)
        
        with self.login("gimli", "gloin"):
            response = self.post("delete_comment", comment_id=comment.pk)
            self.assertEqual(response.status_code, 302)
            self.assertEqual(Comment.objects.count(), 0)
    
    def test_ttag_comment_count(self):
        g = get_user_model().objects.create(username="Sauron")
        self.post_comment(g, data={
            "name": "Gandalf",
            "comment": "You can't win",
        })
        self.post_comment(g, data={
            "name": "Gollum",
            "comment": "We wants our precious",
        })
        
        self.assert_renders(
            "{% load dialogos_tags %}{% comment_count o %}",
            Context({"o": g}),
            "2"
        )
    
    def test_ttag_comments(self):
        g = get_user_model().objects.create(username="Sauron")
        self.post_comment(g, data={
            "name": "Gandalf",
            "comment": "You can't win",
        })
        self.post_comment(g, data={
            "name": "Gollum",
            "comment": "We wants our precious",
        })
        
        c = Context({"o": g})
        self.assert_renders(
            "{% load dialogos_tags %}{% comments o as cs %}",
            c,
            ""
        )
        self.assertEqual(list(c["cs"]), list(Comment.objects.all()))
    
    def test_ttag_comment_form(self):
        g = get_user_model().objects.create(username="Sauron")
        c = Context({"o": g})
        self.assert_renders(
            "{% load dialogos_tags %}{% comment_form o as comment_form %}",
            c,
            ""
        )
        self.assertTrue(isinstance(c["comment_form"], CommentForm))
        
        with self.login("gimli", "gloin"):
            c = Context({"o": g, "user": self.user})
            self.assert_renders(
                "{% load dialogos_tags %}{% comment_form o as comment_form %}",
                c,
                ""
            )
            self.assertTrue(isinstance(c["comment_form"], CommentForm))
    
    def test_ttag_comment_target(self):
        g = get_user_model().objects.create(username="legolas")
        self.assert_renders(
            "{% load dialogos_tags %}{% comment_target o %}",
            Context({"o": g}),
            "/comment/%d/%d/" % (ContentType.objects.get_for_model(g).pk, g.pk)
        )
